@@ -4,7 +4,6 @@ using namespace Rcpp;
 #include "haven_types.h"
 
 ssize_t data_writer(const void *data, size_t len, void *ctx);
-std::string rClass(RObject x);
 
 inline const char* string_utf8(SEXP x, int i) {
   return Rf_translateCharUTF8(STRING_ELT(x, i));
@@ -34,6 +33,17 @@ inline readstat_measure_e measureType(SEXP x) {
   }
 }
 
+inline int displayWidth(RObject x) {
+  RObject display_width_obj = x.attr("display_width");
+  switch(TYPEOF(display_width_obj)) {
+  case INTSXP:
+    return INTEGER(display_width_obj)[0];
+  case REALSXP:
+    return REAL(display_width_obj)[0];
+  }
+  return 0;
+}
+
 
 class Writer {
   FileType type_;
@@ -42,7 +52,9 @@ class Writer {
   FILE* pOut_;
 
 public:
-  Writer(FileType type, List x, std::string path): type_(type), x_(x) {
+  Writer(FileType type, List x, CharacterVector pathEnc): type_(type), x_(x) {
+    std::string path(Rf_translateChar(pathEnc[0]));
+
     pOut_ = fopen(path.c_str(), "wb");
     if (pOut_ == NULL)
       stop("Failed to open '%s' for writing", path);
@@ -58,16 +70,24 @@ public:
     } catch (...) {};
   }
 
+  void setCompression(readstat_compress_t version) {
+    readstat_writer_set_compression(writer_, version);
+  }
+
   void setVersion(int version) {
     readstat_writer_set_file_format_version(writer_, version);
   }
 
+  void setName(const std::string& name) {
+    readstat_writer_set_table_name(writer_, name.c_str());
+  }
+
   void write() {
     int p = x_.size();
     if (p == 0)
       return;
 
-    CharacterVector names = as<CharacterVector>(x_.attr("names"));
+    CharacterVector names(as<CharacterVector>(x_.attr("names")));
 
     // Define variables
     for (int j = 0; j < p; ++j) {
@@ -100,6 +120,9 @@ public:
     case HAVEN_SAS:
       checkStatus(readstat_begin_writing_sas7bdat(writer_, this, n));
       break;
+    case HAVEN_XPT:
+      checkStatus(readstat_begin_writing_xport(writer_, this, n));
+      break;
     }
 
     // Write data
@@ -163,18 +186,21 @@ public:
 
     case HAVEN_DATETIME:
       switch(type_) {
+      case HAVEN_XPT:
       case HAVEN_SAS:   return "DATETIME";
       case HAVEN_SPSS:  return "DATETIME";
       case HAVEN_STATA: return "%tc";
       }
     case HAVEN_DATE:
       switch(type_) {
+      case HAVEN_XPT:
       case HAVEN_SAS:   return "DATE";
       case HAVEN_SPSS:  return "DATE";
       case HAVEN_STATA: return "%td";
       }
     case HAVEN_TIME:
       switch(type_) {
+      case HAVEN_XPT:
       case HAVEN_SAS:   return "TIME";
       case HAVEN_SPSS:  return "TIME";
       case HAVEN_STATA: return NULL; // Stata doesn't have a pure time type
@@ -186,13 +212,13 @@ public:
 
   void defineVariable(IntegerVector x, const char* name, const char* format = NULL) {
     readstat_label_set_t* labelSet = NULL;
-    if (rClass(x) == "factor") {
+    if (Rf_inherits(x, "factor")) {
       labelSet = readstat_add_label_set(writer_, READSTAT_TYPE_INT32, name);
 
       CharacterVector levels = as<CharacterVector>(x.attr("levels"));
       for (int i = 0; i < levels.size(); ++i)
         readstat_label_int32_value(labelSet, i + 1, string_utf8(levels, i));
-    } else if (rClass(x) == "labelled") {
+    } else if (Rf_inherits(x, "labelled")) {
       labelSet = readstat_add_label_set(writer_, READSTAT_TYPE_INT32, name);
 
       IntegerVector values = as<IntegerVector>(x.attr("labels"));
@@ -208,11 +234,12 @@ public:
     readstat_variable_set_label(var, var_label(x));
     readstat_variable_set_label_set(var, labelSet);
     readstat_variable_set_measure(var, measureType(x));
+    readstat_variable_set_display_width(var, displayWidth(x));
   }
 
   void defineVariable(NumericVector x, const char* name, const char* format = NULL) {
     readstat_label_set_t* labelSet = NULL;
-    if (rClass(x) == "labelled") {
+    if (Rf_inherits(x, "labelled")) {
       labelSet = readstat_add_label_set(writer_, READSTAT_TYPE_DOUBLE, name);
 
       NumericVector values = as<NumericVector>(x.attr("labels"));
@@ -224,15 +251,32 @@ public:
 
     readstat_variable_t* var =
       readstat_add_variable(writer_, name, READSTAT_TYPE_DOUBLE, 0);
+
     readstat_variable_set_format(var, format);
     readstat_variable_set_label(var, var_label(x));
     readstat_variable_set_label_set(var, labelSet);
     readstat_variable_set_measure(var, measureType(x));
+    readstat_variable_set_display_width(var, displayWidth(x));
+
+    if (Rf_inherits(x, "labelled_spss")) {
+      SEXP na_range = x.attr("na_range");
+      if (TYPEOF(na_range) == REALSXP && Rf_length(na_range) == 2) {
+        readstat_variable_add_missing_double_range(var, REAL(na_range)[0], REAL(na_range)[1]);
+      }
+
+      SEXP na_values = x.attr("na_values");
+      if (TYPEOF(na_values) == REALSXP) {
+        int n = Rf_length(na_values);
+        for (int i = 0; i < n; ++i) {
+          readstat_variable_add_missing_double_value(var, REAL(na_values)[i]);
+        }
+      }
+    }
   }
 
   void defineVariable(CharacterVector x, const char* name, const char* format = NULL) {
     readstat_label_set_t* labelSet = NULL;
-    if (rClass(x) == "labelled") {
+    if (Rf_inherits(x, "labelled")) {
       labelSet = readstat_add_label_set(writer_, READSTAT_TYPE_STRING, name);
 
       CharacterVector values = as<CharacterVector>(x.attr("labels"));
@@ -241,10 +285,9 @@ public:
       for (int i = 0; i < values.size(); ++i)
         readstat_label_string_value(labelSet, string_utf8(values, i), string_utf8(labels, i));
     }
-
     int max_length = 0;
     for (int i = 0; i < x.size(); ++i) {
-      int length = std::string(x[i]).size();
+      int length = strlen(string_utf8(x, i));
       if (length > max_length)
         max_length = length;
     }
@@ -255,6 +298,7 @@ public:
     readstat_variable_set_label(var, var_label(x));
     readstat_variable_set_label_set(var, labelSet);
     readstat_variable_set_measure(var, measureType(x));
+    readstat_variable_set_display_width(var, displayWidth(x));
   }
 
   // Value helper -------------------------------------------------------------
@@ -300,28 +344,30 @@ ssize_t data_writer(const void *data, size_t len, void *ctx) {
   return ((Writer*) ctx)->write(data, len);
 }
 
-std::string rClass(RObject x) {
-  RObject klass_ = x.attr("class");
-  std::string klass;
-  if (klass_ == R_NilValue)
-    return "";
-
-  CharacterVector klassv = as<Rcpp::CharacterVector>(klass_);
-  return std::string(klassv[0]);
-}
 // [[Rcpp::export]]
-void write_sav_(List data, std::string path) {
-  Writer(HAVEN_SPSS, data, path).write();
+void write_sav_(List data, CharacterVector path, bool compress) {
+  Writer writer(HAVEN_SPSS, data, path);
+  if (compress)
+    writer.setCompression(READSTAT_COMPRESS_BINARY);
+  writer.write();
 }
 
 // [[Rcpp::export]]
-void write_dta_(List data, std::string path, int version) {
+void write_dta_(List data, CharacterVector path, int version) {
   Writer writer(HAVEN_STATA, data, path);
   writer.setVersion(version);
   writer.write();
 }
 
 // [[Rcpp::export]]
-void write_sas_(List data, std::string path) {
+void write_sas_(List data, CharacterVector path) {
   Writer(HAVEN_SAS, data, path).write();
 }
+
+// [[Rcpp::export]]
+void write_xpt_(List data, CharacterVector path, int version, std::string name) {
+  Writer writer(HAVEN_XPT, data, path);
+  writer.setVersion(version);
+  writer.setName(name);
+  writer.write();
+}
