@@ -106,18 +106,30 @@ static readstat_error_t readstat_begin_writing_data(readstat_writer_t *writer) {
         variable->offset = row_len;
         row_len += variable->storage_width;
     }
-    if (writer->callbacks.begin_data) {
-        retval = writer->callbacks.begin_data(writer);
+    if (writer->callbacks.variable_ok) {
+        for (i=0; i<writer->variables_count; i++) {
+            readstat_variable_t *variable = readstat_get_variable(writer, i);
+            retval = writer->callbacks.variable_ok(variable);
+            if (retval != READSTAT_OK)
+                goto cleanup;
+        }
     }
     writer->row_len = row_len;
     writer->row = malloc(writer->row_len);
+    if (writer->callbacks.begin_data) {
+        retval = writer->callbacks.begin_data(writer);
+    }
 
+cleanup:
     return retval;
 }
 
 void readstat_writer_free(readstat_writer_t *writer) {
     int i;
     if (writer) {
+        if (writer->callbacks.module_ctx_free && writer->module_ctx) {
+            writer->callbacks.module_ctx_free(writer->module_ctx);
+        }
         if (writer->variables) {
             for (i=0; i<writer->variables_count; i++) {
                 readstat_variable_free(writer->variables[i]);
@@ -264,6 +276,12 @@ readstat_label_set_t *readstat_get_label_set(readstat_writer_t *writer, int inde
     return NULL;
 }
 
+void readstat_sort_label_set(readstat_label_set_t *label_set,
+        int (*compare)(const readstat_value_label_t *, const readstat_value_label_t *)) {
+    qsort(label_set->value_labels, label_set->value_labels_count, sizeof(readstat_value_label_t),
+            (int (*)(const void *, const void *))compare);
+}
+
 readstat_value_label_t *readstat_get_value_label(readstat_label_set_t *label_set, int index) {
     if (index < label_set->value_labels_count) {
         return &label_set->value_labels[index];
@@ -423,6 +441,11 @@ readstat_error_t readstat_writer_set_file_timestamp(readstat_writer_t *writer, t
     return READSTAT_OK;
 }
 
+readstat_error_t readstat_writer_set_table_name(readstat_writer_t *writer, const char *table_name) {
+    snprintf(writer->table_name, sizeof(writer->table_name), "%s", table_name);
+    return READSTAT_OK;
+}
+
 readstat_error_t readstat_writer_set_fweight_variable(readstat_writer_t *writer, const readstat_variable_t *variable) {
     if (readstat_variable_get_type_class(variable) == READSTAT_TYPE_CLASS_STRING)
         return READSTAT_ERROR_BAD_FREQUENCY_WEIGHT;
@@ -431,7 +454,7 @@ readstat_error_t readstat_writer_set_fweight_variable(readstat_writer_t *writer,
     return READSTAT_OK;
 }
 
-readstat_error_t readstat_writer_set_file_format_version(readstat_writer_t *writer, long version) {
+readstat_error_t readstat_writer_set_file_format_version(readstat_writer_t *writer, uint8_t version) {
     writer->version = version;
     return READSTAT_OK;
 }
