@@ -5,6 +5,7 @@
 #include <string.h>
 #include <math.h>
 #include <time.h>
+#include <limits.h>
 #include <inttypes.h>
 
 #include "readstat_sas.h"
@@ -12,8 +13,9 @@
 #include "../readstat_convert.h"
 #include "../readstat_writer.h"
 
-#define HEADER_SIZE 1024
-#define PAGE_SIZE   4096
+#define SAS_FILE_HEADER_SIZE_32BIT 1024
+#define SAS_FILE_HEADER_SIZE_64BIT 8192
+#define SAS_DEFAULT_PAGE_SIZE      4096
 
 #define SAS_DEFAULT_STRING_ENCODING "WINDOWS-1252"
 
@@ -59,6 +61,7 @@ static readstat_charset_entry_t _charset_table[] = {
     { .code = 65,    .name = "WINDOWS-1255" },
     { .code = 66,    .name = "WINDOWS-1256" },
     { .code = 67,    .name = "WINDOWS-1257" },
+    { .code = 68,    .name = "WINDOWS-1258" },
     { .code = 119,   .name = "EUC-TW" },
     { .code = 123,   .name = "BIG-5" },
     { .code = 125,   .name = "GB18030" }, // "euc-cn" in SAS
@@ -67,6 +70,17 @@ static readstat_charset_entry_t _charset_table[] = {
     { .code = 140,   .name = "EUC-KR" }
 };
 
+static time_t sas_convert_time(double time, time_t epoch) {
+    time += epoch;
+    if (isnan(time))
+        return 0;
+    if (time > 1.0 * LONG_MAX)
+        return LONG_MAX;
+    if (time < 1.0 * LONG_MIN)
+        return LONG_MIN;
+    return time;
+}
+
 uint64_t sas_read8(const char *data, int bswap) {
     uint64_t tmp;
     memcpy(&tmp, data, 8);
@@ -85,6 +99,10 @@ uint16_t sas_read2(const char *data, int bswap) {
     return bswap ? byteswap2(tmp) : tmp;
 }
 
+size_t sas_subheader_remainder(size_t len, size_t signature_len) {
+    return len - (4+2*signature_len);
+}
+
 readstat_error_t sas_read_header(readstat_io_t *io, sas_header_info_t *hinfo, 
         readstat_error_handler error_handler, void *user_ctx) {
     sas_header_start_t  header_start;
@@ -129,7 +147,7 @@ readstat_error_t sas_read_header(readstat_io_t *io, sas_header_info_t *hinfo,
     }
     if (hinfo->encoding == NULL) {
         if (error_handler) {
-            snprintf(error_buf, sizeof(error_buf), "Unsupported character set code: %d\n", header_start.encoding);
+            snprintf(error_buf, sizeof(error_buf), "Unsupported character set code: %d", header_start.encoding);
             error_handler(error_buf, user_ctx);
         }
         retval = READSTAT_ERROR_UNSUPPORTED_CHARSET;
@@ -142,16 +160,23 @@ readstat_error_t sas_read_header(readstat_io_t *io, sas_header_info_t *hinfo,
     }
 
     double creation_time, modification_time;
+
     if (io->read(&creation_time, sizeof(double), io->io_ctx) < sizeof(double)) {
         retval = READSTAT_ERROR_READ;
         goto cleanup;
     }
+    if (bswap)
+        creation_time = byteswap_double(creation_time);
+
     if (io->read(&modification_time, sizeof(double), io->io_ctx) < sizeof(double)) {
         retval = READSTAT_ERROR_READ;
         goto cleanup;
     }
-    hinfo->creation_time = bswap ? byteswap_double(creation_time) + epoch : creation_time + epoch;
-    hinfo->modification_time = bswap ? byteswap_double(creation_time) + epoch : creation_time + epoch;
+    if (bswap)
+        modification_time = byteswap_double(modification_time);
+
+    hinfo->creation_time = sas_convert_time(creation_time, epoch);
+    hinfo->modification_time = sas_convert_time(modification_time, epoch);
 
     if (io->seek(16, READSTAT_SEEK_CUR, io->io_ctx) == -1) {
         retval = READSTAT_ERROR_SEEK;
@@ -170,13 +195,16 @@ readstat_error_t sas_read_header(readstat_io_t *io, sas_header_info_t *hinfo,
     }
 
     hinfo->header_size = bswap ? byteswap4(header_size) : header_size;
+    hinfo->page_size = bswap ? byteswap4(page_size) : page_size;
 
-    if (hinfo->header_size < 1024) {
+    if (hinfo->header_size < 1024 || hinfo->page_size < 1024) {
+        retval = READSTAT_ERROR_PARSE;
+        goto cleanup;
+    }
+    if (hinfo->header_size > (1<<20) || hinfo->page_size > (1<<24)) {
         retval = READSTAT_ERROR_PARSE;
         goto cleanup;
     }
-
-    hinfo->page_size = bswap ? byteswap4(page_size) : page_size;
 
     if (hinfo->u64) {
         hinfo->page_header_size = SAS_PAGE_HEADER_SIZE_64BIT;
@@ -201,11 +229,15 @@ readstat_error_t sas_read_header(readstat_io_t *io, sas_header_info_t *hinfo,
         }
         hinfo->page_count = bswap ? byteswap4(page_count) : page_count;
     }
+    if (hinfo->page_count > (1<<24)) {
+        retval = READSTAT_ERROR_PARSE;
+        goto cleanup;
+    }
     
     if (io->seek(8, READSTAT_SEEK_CUR, io->io_ctx) == -1) {
         retval = READSTAT_ERROR_SEEK;
         if (error_handler) {
-            snprintf(error_buf, sizeof(error_buf), "ReadStat: Failed to seek forward by %d\n", 8);
+            snprintf(error_buf, sizeof(error_buf), "ReadStat: Failed to seek forward by %d", 8);
             error_handler(error_buf, user_ctx);
         }
         goto cleanup;
@@ -230,7 +262,7 @@ readstat_error_t sas_read_header(readstat_io_t *io, sas_header_info_t *hinfo,
         retval = READSTAT_ERROR_SEEK;
         if (error_handler) {
             snprintf(error_buf, sizeof(error_buf), 
-                    "ReadStat: Failed to seek to position %" PRId64 "\n", hinfo->header_size);
+                    "ReadStat: Failed to seek to position %" PRId64, hinfo->header_size);
             error_handler(error_buf, user_ctx);
         }
         goto cleanup;
@@ -300,9 +332,9 @@ readstat_error_t sas_write_header(readstat_writer_t *writer, sas_header_info_t *
     if (retval != READSTAT_OK)
         goto cleanup;
 
-    snprintf(header_end.release, sizeof(header_end.release),
-            "%1ld.%04ldM0", writer->version / 10000, writer->version % 10000);
-    header_end.release[sizeof(header_end.release)-1] = '0';
+    char release[32];
+    snprintf(release, sizeof(release), "%1ld.%04dM0", writer->version, 101);
+    strncpy(header_end.release, release, sizeof(header_end.release));
 
     retval = readstat_write_bytes(writer, &header_end, sizeof(sas_header_end_t));
     if (retval != READSTAT_OK)
@@ -320,14 +352,15 @@ sas_header_info_t *sas_header_info_init(readstat_writer_t *writer, int is_64bit)
     sas_header_info_t *hinfo = calloc(1, sizeof(sas_header_info_t));
     hinfo->creation_time = writer->timestamp;
     hinfo->modification_time = writer->timestamp;
-    hinfo->header_size = HEADER_SIZE;
-    hinfo->page_size = PAGE_SIZE;
+    hinfo->page_size = SAS_DEFAULT_PAGE_SIZE;
     hinfo->u64 = !!is_64bit;
 
     if (hinfo->u64) {
+        hinfo->header_size = SAS_FILE_HEADER_SIZE_64BIT;
         hinfo->page_header_size = SAS_PAGE_HEADER_SIZE_64BIT;
         hinfo->subheader_pointer_size = SAS_SUBHEADER_POINTER_SIZE_64BIT;
     } else {
+        hinfo->header_size = SAS_FILE_HEADER_SIZE_32BIT;
         hinfo->page_header_size = SAS_PAGE_HEADER_SIZE_32BIT;
         hinfo->subheader_pointer_size = SAS_SUBHEADER_POINTER_SIZE_32BIT;
     }
@@ -344,7 +377,7 @@ readstat_error_t sas_fill_page(readstat_writer_t *writer, sas_header_info_t *hin
     return READSTAT_OK;
 }
 
-readstat_error_t sas_validate_name(const char *name) {
+readstat_error_t sas_validate_name(const char *name, size_t max_len) {
     int j;
     for (j=0; name[j]; j++) {
         if (name[j] != '_' &&
@@ -366,20 +399,12 @@ readstat_error_t sas_validate_name(const char *name) {
         return READSTAT_ERROR_NAME_IS_RESERVED_WORD;
     }
 
-    if (strlen(name) > 32)
+    if (strlen(name) > max_len)
         return READSTAT_ERROR_NAME_IS_TOO_LONG;
 
     return READSTAT_OK;
 }
 
-readstat_error_t sas_validate_column_names(readstat_writer_t *writer) {
-    int i;
-    for (i=0; i<writer->variables_count; i++) {
-        readstat_variable_t *variable = readstat_get_variable(writer, i);
-        readstat_error_t error = sas_validate_name(readstat_variable_get_name(variable));
-        if (error != READSTAT_OK)
-            return error;
-    }
-    return READSTAT_OK;
+readstat_error_t sas_validate_variable(readstat_variable_t *variable) {
+    return sas_validate_name(readstat_variable_get_name(variable), 32);
 }
-
