@@ -19,20 +19,15 @@ typedef struct col_info_s {
     sas_text_ref_t  format_ref;
     sas_text_ref_t  label_ref;
 
-    int    index;
-    int    offset;
-    int    width;
-    int    type;
+    int         index;
+    uint64_t    offset;
+    uint32_t    width;
+    int         type;
 } col_info_t;
 
 typedef struct sas7bdat_ctx_s {
-    readstat_info_handler       info_handler;
-    readstat_metadata_handler   metadata_handler;
-    readstat_variable_handler   variable_handler;
-    readstat_value_handler      value_handler;
-    readstat_error_handler      error_handler;
-    readstat_progress_handler   progress_handler;
-    int64_t                     file_size;
+    readstat_callbacks_t handle;
+    int64_t              file_size;
 
     int            little_endian;
     int            u64;
@@ -52,6 +47,7 @@ typedef struct sas7bdat_ctx_s {
     uint64_t        page_count;
     uint64_t        page_size;
     char           *page;
+    char           *row;
 
     uint64_t        page_header_size;
     uint64_t        subheader_pointer_size;
@@ -77,7 +73,8 @@ typedef struct sas7bdat_ctx_s {
     const char    *output_encoding;
     iconv_t        converter;
 
-    time_t         timestamp;
+    time_t         ctime;
+    time_t         mtime;
     int            version;
     char           file_label[4*64+1];
     char           error_buf[2048];
@@ -108,6 +105,9 @@ static void sas7bdat_ctx_free(sas7bdat_ctx_t *ctx) {
     if (ctx->page)
         free(ctx->page);
 
+    if (ctx->row)
+        free(ctx->row);
+
     if (ctx->converter)
         iconv_close(ctx->converter);
 
@@ -116,7 +116,7 @@ static void sas7bdat_ctx_free(sas7bdat_ctx_t *ctx) {
 
 static readstat_error_t sas7bdat_update_progress(sas7bdat_ctx_t *ctx) {
     readstat_io_t *io = ctx->io;
-    return io->update(ctx->file_size, ctx->progress_handler, ctx->user_ctx, io->io_ctx);
+    return io->update(ctx->file_size, ctx->handle.progress, ctx->user_ctx, io->io_ctx);
 }
 
 static readstat_error_t sas7bdat_parse_column_text_subheader(const char *subheader, size_t len, sas7bdat_ctx_t *ctx) {
@@ -124,7 +124,7 @@ static readstat_error_t sas7bdat_parse_column_text_subheader(const char *subhead
     size_t signature_len = ctx->u64 ? 8 : 4;
     uint16_t remainder = sas_read2(&subheader[signature_len], ctx->bswap);
     char *blob = NULL;
-    if (remainder != len - (4+2*signature_len)) {
+    if (remainder != sas_subheader_remainder(len, signature_len)) {
         retval = READSTAT_ERROR_PARSE;
         goto cleanup;
     }
@@ -217,6 +217,12 @@ static readstat_error_t sas7bdat_parse_row_size_subheader(const char *subheader,
     }
 
     ctx->row_length = row_length;
+    ctx->row = readstat_realloc(ctx->row, ctx->row_length);
+    if (ctx->row == NULL) {
+        retval = READSTAT_ERROR_MALLOC;
+        goto cleanup;
+    }
+
     ctx->page_row_count = page_row_count;
     if (ctx->row_limit == 0 || total_row_count < ctx->row_limit)
         ctx->row_limit = total_row_count;
@@ -261,7 +267,7 @@ static readstat_error_t sas7bdat_parse_column_name_subheader(const char *subhead
     const char *cnp = &subheader[signature_len+8];
     uint16_t remainder = sas_read2(&subheader[signature_len], ctx->bswap);
 
-    if (remainder != len - (4+2*signature_len)) {
+    if (remainder != sas_subheader_remainder(len, signature_len)) {
         retval = READSTAT_ERROR_PARSE;
         goto cleanup;
     }
@@ -289,7 +295,7 @@ static readstat_error_t sas7bdat_parse_column_attributes_subheader(const char *s
     const char *cap = &subheader[signature_len+8];
     uint16_t remainder = sas_read2(&subheader[signature_len], ctx->bswap);
 
-    if (remainder != len - (4+2*signature_len)) {
+    if (remainder != sas_subheader_remainder(len, signature_len)) {
         retval = READSTAT_ERROR_PARSE;
         goto cleanup;
     }
@@ -320,7 +326,7 @@ static readstat_error_t sas7bdat_parse_column_attributes_subheader(const char *s
             }
         } else if (cap[off+6] == SAS_COLUMN_TYPE_CHR) {
             ctx->col_info[i].type = READSTAT_TYPE_STRING;
-            if (ctx->col_info[i].width < 0) {
+            if (ctx->col_info[i].width > INT16_MAX || ctx->col_info[i].width == 0) {
                 retval = READSTAT_ERROR_PARSE;
                 goto cleanup;
             }
@@ -371,11 +377,11 @@ static readstat_error_t sas7bdat_handle_data_value(readstat_variable_t *variable
         retval = readstat_convert(ctx->scratch_buffer, ctx->scratch_buffer_len,
                 col_data, col_info->width, ctx->converter);
         if (retval != READSTAT_OK) {
-            if (ctx->error_handler) {
+            if (ctx->handle.error) {
                 snprintf(ctx->error_buf, sizeof(ctx->error_buf),
                         "ReadStat: Error converting string to specified encoding: %.*s",
                         col_info->width, col_data);
-                ctx->error_handler(ctx->error_buf, ctx->user_ctx);
+                ctx->handle.error(ctx->error_buf, ctx->user_ctx);
             }
             goto cleanup;
         }
@@ -411,7 +417,7 @@ static readstat_error_t sas7bdat_handle_data_value(readstat_variable_t *variable
             value.v.double_value = dval;
         }
     }
-    cb_retval = ctx->value_handler(ctx->parsed_row_count, variable, value, ctx->user_ctx);
+    cb_retval = ctx->handle.value(ctx->parsed_row_count, variable, value, ctx->user_ctx);
 
     if (cb_retval != READSTAT_HANDLER_OK)
         retval = READSTAT_ERROR_USER_ABORT;
@@ -426,7 +432,7 @@ static readstat_error_t sas7bdat_parse_single_row(const char *data, sas7bdat_ctx
 
     readstat_error_t retval = READSTAT_OK;
     int j;
-    if (ctx->value_handler) {
+    if (ctx->handle.value) {
         ctx->scratch_buffer_len = 4*ctx->max_col_width+1;
         ctx->scratch_buffer = readstat_realloc(ctx->scratch_buffer, ctx->scratch_buffer_len);
         if (ctx->scratch_buffer == NULL) {
@@ -440,7 +446,7 @@ static readstat_error_t sas7bdat_parse_single_row(const char *data, sas7bdat_ctx
             if (variable->skip)
                 continue;
 
-            if (col_info->offset < 0 || col_info->offset + col_info->width > ctx->row_length) {
+            if (col_info->offset > ctx->row_length || col_info->offset + col_info->width > ctx->row_length) {
                 retval = READSTAT_ERROR_PARSE;
                 goto cleanup;
             }
@@ -480,32 +486,23 @@ static readstat_error_t sas7bdat_parse_subheader_rle(const char *subheader, size
         return READSTAT_OK;
 
     readstat_error_t retval = READSTAT_OK;
-    char *buffer = NULL;
     ssize_t bytes_decompressed = 0;
 
-    if ((buffer = readstat_malloc(ctx->row_length)) == NULL) {
-        retval = READSTAT_ERROR_MALLOC;
-        goto cleanup;
-    }
-
-    bytes_decompressed = sas_rle_decompress(
-            buffer, ctx->row_length, subheader, len);
+    bytes_decompressed = sas_rle_decompress(ctx->row, ctx->row_length, subheader, len);
 
     if (bytes_decompressed != ctx->row_length) {
         retval = READSTAT_ERROR_ROW_WIDTH_MISMATCH;
-        if (ctx->error_handler) {
+        if (ctx->handle.error) {
             snprintf(ctx->error_buf, sizeof(ctx->error_buf), 
                     "ReadStat: Row #%d decompressed to %ld bytes (expected %d bytes)",
                     ctx->parsed_row_count, (long)(bytes_decompressed), ctx->row_length);
-            ctx->error_handler(ctx->error_buf, ctx->user_ctx);
+            ctx->handle.error(ctx->error_buf, ctx->user_ctx);
         }
         goto cleanup;
     }
-    retval = sas7bdat_parse_single_row(buffer, ctx);
-cleanup:
-    if (buffer)
-        free(buffer);
+    retval = sas7bdat_parse_single_row(ctx->row, ctx);
 
+cleanup:
     return retval;
 }
 
@@ -573,11 +570,11 @@ cleanup:
             *out_retval = retval;
 
         if (retval == READSTAT_ERROR_CONVERT_BAD_STRING) {
-            if (ctx->error_handler) {
+            if (ctx->handle.error) {
                 snprintf(ctx->error_buf, sizeof(ctx->error_buf),
                         "ReadStat: Error converting variable #%d info to specified encoding: %s %s (%s)",
                         i, variable->name, variable->format, variable->label);
-                ctx->error_handler(ctx->error_buf, ctx->user_ctx);
+                ctx->handle.error(ctx->error_buf, ctx->user_ctx);
             }
         }
 
@@ -587,16 +584,22 @@ cleanup:
     return variable;
 }
 
-static readstat_error_t sas7bdat_submit_columns(sas7bdat_ctx_t *ctx) {
+static readstat_error_t sas7bdat_submit_columns(sas7bdat_ctx_t *ctx, int compressed) {
     readstat_error_t retval = READSTAT_OK;
-    if (ctx->info_handler) {
-        if (ctx->info_handler(ctx->row_limit, ctx->column_count, ctx->user_ctx) != READSTAT_HANDLER_OK) {
-            retval = READSTAT_ERROR_USER_ABORT;
-            goto cleanup;
-        }
-    }
-    if (ctx->metadata_handler) {
-        if (ctx->metadata_handler(ctx->file_label, ctx->input_encoding, ctx->timestamp, ctx->version, ctx->user_ctx) != READSTAT_HANDLER_OK) {
+    if (ctx->handle.metadata) {
+        readstat_metadata_t metadata = {
+            .row_count = ctx->row_limit,
+            .var_count = ctx->column_count,
+            .file_label = ctx->file_label,
+            .file_encoding = ctx->input_encoding, /* orig encoding? */
+            .creation_time = ctx->ctime,
+            .modified_time = ctx->mtime,
+            .file_format_version = ctx->version,
+            .compression = compressed ? READSTAT_COMPRESS_ROWS : READSTAT_COMPRESS_NONE,
+            .endianness = ctx->little_endian ? READSTAT_ENDIAN_LITTLE : READSTAT_ENDIAN_BIG,
+            .is64bit = ctx->u64
+        };
+        if (ctx->handle.metadata(&metadata, ctx->user_ctx) != READSTAT_HANDLER_OK) {
             retval = READSTAT_ERROR_USER_ABORT;
             goto cleanup;
         }
@@ -614,8 +617,8 @@ static readstat_error_t sas7bdat_submit_columns(sas7bdat_ctx_t *ctx) {
             break;
 
         int cb_retval = READSTAT_HANDLER_OK;
-        if (ctx->variable_handler) {
-            cb_retval = ctx->variable_handler(i, ctx->variables[i], ctx->variables[i]->format, ctx->user_ctx);
+        if (ctx->handle.variable) {
+            cb_retval = ctx->handle.variable(i, ctx->variables[i], ctx->variables[i]->format, ctx->user_ctx);
         }
         if (cb_retval == READSTAT_HANDLER_ABORT) {
             retval = READSTAT_ERROR_USER_ABORT;
@@ -632,10 +635,10 @@ cleanup:
     return retval;
 }
 
-static readstat_error_t sas7bdat_submit_columns_if_needed(sas7bdat_ctx_t *ctx) {
+static readstat_error_t sas7bdat_submit_columns_if_needed(sas7bdat_ctx_t *ctx, int compressed) {
     readstat_error_t retval = READSTAT_OK;
     if (!ctx->did_submit_columns) {
-        if ((retval = sas7bdat_submit_columns(ctx)) != READSTAT_OK) {
+        if ((retval = sas7bdat_submit_columns(ctx, compressed)) != READSTAT_OK) {
             goto cleanup;
         }
         ctx->did_submit_columns = 1;
@@ -768,7 +771,7 @@ static readstat_error_t sas7bdat_parse_page_pass2(const char *page, size_t page_
                             retval = READSTAT_ERROR_ROW_WIDTH_MISMATCH;
                             goto cleanup;
                         }
-                        if ((retval = sas7bdat_submit_columns_if_needed(ctx)) != READSTAT_OK) {
+                        if ((retval = sas7bdat_submit_columns_if_needed(ctx, 1)) != READSTAT_OK) {
                             goto cleanup;
                         }
                         if ((retval = sas7bdat_parse_single_row(page + offset, ctx)) != READSTAT_OK) {
@@ -782,7 +785,7 @@ static readstat_error_t sas7bdat_parse_page_pass2(const char *page, size_t page_
                         }
                     }
                 } else if (compression == SAS_COMPRESSION_ROW) {
-                    if ((retval = sas7bdat_submit_columns_if_needed(ctx)) != READSTAT_OK) {
+                    if ((retval = sas7bdat_submit_columns_if_needed(ctx, 1)) != READSTAT_OK) {
                         goto cleanup;
                     }
                     if ((retval = sas7bdat_parse_subheader_rle(page + offset, len, ctx)) != READSTAT_OK) {
@@ -813,10 +816,10 @@ static readstat_error_t sas7bdat_parse_page_pass2(const char *page, size_t page_
         }
     }
     if (data) {
-        if ((retval = sas7bdat_submit_columns_if_needed(ctx)) != READSTAT_OK) {
+        if ((retval = sas7bdat_submit_columns_if_needed(ctx, 0)) != READSTAT_OK) {
             goto cleanup;
         }
-        if (ctx->value_handler) {
+        if (ctx->handle.value) {
             retval = sas7bdat_parse_rows(data, page + page_size - data, ctx);
         }
     } 
@@ -834,11 +837,11 @@ static readstat_error_t sas7bdat_parse_meta_pages_pass1(sas7bdat_ctx_t *ctx, int
     for (i=0; i<ctx->page_count; i++) {
         if (io->seek(ctx->header_size + i*ctx->page_size, READSTAT_SEEK_SET, io->io_ctx) == -1) {
             retval = READSTAT_ERROR_SEEK;
-            if (ctx->error_handler) {
+            if (ctx->handle.error) {
                 snprintf(ctx->error_buf, sizeof(ctx->error_buf), "ReadStat: Failed to seek to position %" PRId64 
                         " (= %" PRId64 " + %" PRId64 "*%" PRId64 ")",
                         ctx->header_size + i*ctx->page_size, ctx->header_size, i, ctx->page_size);
-                ctx->error_handler(ctx->error_buf, ctx->user_ctx);
+                ctx->handle.error(ctx->error_buf, ctx->user_ctx);
             }
             goto cleanup;
         }
@@ -868,12 +871,12 @@ static readstat_error_t sas7bdat_parse_meta_pages_pass1(sas7bdat_ctx_t *ctx, int
         }
 
         if ((retval = sas7bdat_parse_page_pass1(ctx->page, ctx->page_size, ctx)) != READSTAT_OK) {
-            if (ctx->error_handler && retval != READSTAT_ERROR_USER_ABORT) {
+            if (ctx->handle.error && retval != READSTAT_ERROR_USER_ABORT) {
                 int64_t pos = io->seek(0, READSTAT_SEEK_CUR, io->io_ctx);
                 snprintf(ctx->error_buf, sizeof(ctx->error_buf), 
                         "ReadStat: Error parsing page %" PRId64 ", bytes %" PRId64 "-%" PRId64, 
                         i, pos - ctx->page_size, pos-1);
-                ctx->error_handler(ctx->error_buf, ctx->user_ctx);
+                ctx->handle.error(ctx->error_buf, ctx->user_ctx);
             }
             goto cleanup;
         }
@@ -896,11 +899,11 @@ static readstat_error_t sas7bdat_parse_amd_pages_pass1(int64_t last_examined_pag
     for (i=ctx->page_count-1; i>last_examined_page_pass1; i--) {
         if (io->seek(ctx->header_size + i*ctx->page_size, READSTAT_SEEK_SET, io->io_ctx) == -1) {
             retval = READSTAT_ERROR_SEEK;
-            if (ctx->error_handler) {
+            if (ctx->handle.error) {
                 snprintf(ctx->error_buf, sizeof(ctx->error_buf), "ReadStat: Failed to seek to position %" PRId64 
                         " (= %" PRId64 " + %" PRId64 "*%" PRId64 ")",
                         ctx->header_size + i*ctx->page_size, ctx->header_size, i, ctx->page_size);
-                ctx->error_handler(ctx->error_buf, ctx->user_ctx);
+                ctx->handle.error(ctx->error_buf, ctx->user_ctx);
             }
             goto cleanup;
         }
@@ -934,12 +937,12 @@ static readstat_error_t sas7bdat_parse_amd_pages_pass1(int64_t last_examined_pag
         }
 
         if ((retval = sas7bdat_parse_page_pass1(ctx->page, ctx->page_size, ctx)) != READSTAT_OK) {
-            if (ctx->error_handler && retval != READSTAT_ERROR_USER_ABORT) {
+            if (ctx->handle.error && retval != READSTAT_ERROR_USER_ABORT) {
                 int64_t pos = io->seek(0, READSTAT_SEEK_CUR, io->io_ctx);
                 snprintf(ctx->error_buf, sizeof(ctx->error_buf), 
                         "ReadStat: Error parsing page %" PRId64 ", bytes %" PRId64 "-%" PRId64, 
                         i, pos - ctx->page_size, pos-1);
-                ctx->error_handler(ctx->error_buf, ctx->user_ctx);
+                ctx->handle.error(ctx->error_buf, ctx->user_ctx);
             }
             goto cleanup;
         }
@@ -967,12 +970,12 @@ static readstat_error_t sas7bdat_parse_all_pages_pass2(sas7bdat_ctx_t *ctx) {
         }
 
         if ((retval = sas7bdat_parse_page_pass2(ctx->page, ctx->page_size, ctx)) != READSTAT_OK) {
-            if (ctx->error_handler && retval != READSTAT_ERROR_USER_ABORT) {
+            if (ctx->handle.error && retval != READSTAT_ERROR_USER_ABORT) {
                 int64_t pos = io->seek(0, READSTAT_SEEK_CUR, io->io_ctx);
                 snprintf(ctx->error_buf, sizeof(ctx->error_buf), 
                         "ReadStat: Error parsing page %" PRId64 ", bytes %" PRId64 "-%" PRId64, 
                         i, pos - ctx->page_size, pos-1);
-                ctx->error_handler(ctx->error_buf, ctx->user_ctx);
+                ctx->handle.error(ctx->error_buf, ctx->user_ctx);
             }
             goto cleanup;
         }
@@ -992,12 +995,7 @@ readstat_error_t readstat_parse_sas7bdat(readstat_parser_t *parser, const char *
     sas7bdat_ctx_t  *ctx = calloc(1, sizeof(sas7bdat_ctx_t));
     sas_header_info_t  *hinfo = calloc(1, sizeof(sas_header_info_t));
 
-    ctx->info_handler = parser->info_handler;
-    ctx->metadata_handler = parser->metadata_handler;
-    ctx->variable_handler = parser->variable_handler;
-    ctx->value_handler = parser->value_handler;
-    ctx->error_handler = parser->error_handler;
-    ctx->progress_handler = parser->progress_handler;
+    ctx->handle = parser->handlers;
     ctx->input_encoding = parser->input_encoding;
     ctx->output_encoding = parser->output_encoding;
     ctx->user_ctx = user_ctx;
@@ -1011,23 +1009,23 @@ readstat_error_t readstat_parse_sas7bdat(readstat_parser_t *parser, const char *
 
     if ((ctx->file_size = io->seek(0, READSTAT_SEEK_END, io->io_ctx)) == -1) {
         retval = READSTAT_ERROR_SEEK;
-        if (ctx->error_handler) {
+        if (ctx->handle.error) {
             snprintf(ctx->error_buf, sizeof(ctx->error_buf), "ReadStat: Failed to seek to end of file");
-            ctx->error_handler(ctx->error_buf, ctx->user_ctx);
+            ctx->handle.error(ctx->error_buf, ctx->user_ctx);
         }
         goto cleanup;
     }
 
     if (io->seek(0, READSTAT_SEEK_SET, io->io_ctx) == -1) {
         retval = READSTAT_ERROR_SEEK;
-        if (ctx->error_handler) {
+        if (ctx->handle.error) {
             snprintf(ctx->error_buf, sizeof(ctx->error_buf), "ReadStat: Failed to seek to beginning of file");
-            ctx->error_handler(ctx->error_buf, ctx->user_ctx);
+            ctx->handle.error(ctx->error_buf, ctx->user_ctx);
         }
         goto cleanup;
     }
 
-    if ((retval = sas_read_header(io, hinfo, ctx->error_handler, user_ctx)) != READSTAT_OK) {
+    if ((retval = sas_read_header(io, hinfo, ctx->handle.error, user_ctx)) != READSTAT_OK) {
         goto cleanup;
     }
 
@@ -1040,8 +1038,9 @@ readstat_error_t readstat_parse_sas7bdat(readstat_parser_t *parser, const char *
     ctx->page_size = hinfo->page_size;
     ctx->page_header_size = hinfo->page_header_size;
     ctx->subheader_pointer_size = hinfo->subheader_pointer_size;
-    ctx->timestamp = hinfo->modification_time;
-    ctx->version = 10000 * hinfo->major_version + hinfo->minor_version;
+    ctx->ctime = hinfo->creation_time;
+    ctx->mtime = hinfo->modification_time;
+    ctx->version = hinfo->major_version;
     if (ctx->input_encoding == NULL) {
         ctx->input_encoding = hinfo->encoding;
     }
@@ -1074,10 +1073,10 @@ readstat_error_t readstat_parse_sas7bdat(readstat_parser_t *parser, const char *
 
     if (io->seek(ctx->header_size, READSTAT_SEEK_SET, io->io_ctx) == -1) {
         retval = READSTAT_ERROR_SEEK;
-        if (ctx->error_handler) {
+        if (ctx->handle.error) {
             snprintf(ctx->error_buf, sizeof(ctx->error_buf), "ReadStat: Failed to seek to position %" PRId64, 
                     ctx->header_size);
-            ctx->error_handler(ctx->error_buf, ctx->user_ctx);
+            ctx->handle.error(ctx->error_buf, ctx->user_ctx);
         }
         goto cleanup;
     }
@@ -1086,16 +1085,16 @@ readstat_error_t readstat_parse_sas7bdat(readstat_parser_t *parser, const char *
         goto cleanup;
     }
     
-    if ((retval = sas7bdat_submit_columns_if_needed(ctx)) != READSTAT_OK) {
+    if ((retval = sas7bdat_submit_columns_if_needed(ctx, 0)) != READSTAT_OK) {
         goto cleanup;
     }
 
-    if (ctx->value_handler && ctx->parsed_row_count != ctx->row_limit) {
+    if (ctx->handle.value && ctx->parsed_row_count != ctx->row_limit) {
         retval = READSTAT_ERROR_ROW_COUNT_MISMATCH;
-        if (ctx->error_handler) {
+        if (ctx->handle.error) {
             snprintf(ctx->error_buf, sizeof(ctx->error_buf), "ReadStat: Expected %d rows in file, found %d",
                     ctx->row_limit, ctx->parsed_row_count);
-            ctx->error_handler(ctx->error_buf, ctx->user_ctx);
+            ctx->handle.error(ctx->error_buf, ctx->user_ctx);
         }
         goto cleanup;
     }
@@ -1110,10 +1109,10 @@ cleanup:
     if (retval == READSTAT_ERROR_OPEN ||
             retval == READSTAT_ERROR_READ ||
             retval == READSTAT_ERROR_SEEK) {
-        if (ctx->error_handler) {
+        if (ctx->handle.error) {
             snprintf(ctx->error_buf, sizeof(ctx->error_buf), "ReadStat: %s (retval = %d): %s (errno = %d)", 
                     readstat_error_message(retval), retval, strerror(errno), errno);
-            ctx->error_handler(ctx->error_buf, user_ctx);
+            ctx->handle.error(ctx->error_buf, user_ctx);
         }
     }
 
