@@ -3,17 +3,9 @@ extern int8_t   por_ascii_lookup[256];
 extern uint16_t por_unicode_lookup[256];
 
 typedef struct por_ctx_s {
-    readstat_info_handler           info_handler;
-    readstat_metadata_handler       metadata_handler;
-    readstat_variable_handler       variable_handler;
-    readstat_note_handler           note_handler;
-    readstat_fweight_handler        fweight_handler;
-    readstat_value_handler          value_handler;
-    readstat_value_label_handler    value_label_handler;
-    readstat_error_handler          error_handler;
-    readstat_progress_handler       progress_handler;
-    size_t                          file_size;
-    void                           *user_ctx;
+    readstat_callbacks_t    handle;
+    size_t                  file_size;
+    void                   *user_ctx;
 
     int            pos;
     readstat_io_t *io;
@@ -22,6 +14,7 @@ typedef struct por_ctx_s {
     time_t         timestamp;
     long           version;
     char           fweight_name[9];
+    char           file_label[21];
     uint16_t       byte2unicode[256];
     size_t         base30_precision;
     iconv_t        converter;
