@@ -4,6 +4,7 @@
 #include <stdlib.h>
 #include "../readstat.h"
 #include "../readstat_iconv.h"
+#include "../readstat_malloc.h"
 
 #include "readstat_sav.h"
 #include "readstat_sav_parse.h"
@@ -30,7 +31,7 @@ static int count_vars(sav_ctx_t *ctx) {
     spss_varinfo_t *last_info = NULL;
     int var_count = 0;
     for (i=0; i<ctx->var_index; i++) {
-        spss_varinfo_t *info = &ctx->varinfo[i];
+        spss_varinfo_t *info = ctx->varinfo[i];
         if (last_info == NULL || strcmp(info->name, last_info->name) != 0) {
             var_count++;
         }
@@ -40,12 +41,12 @@ static int count_vars(sav_ctx_t *ctx) {
 }
 
 static varlookup_t *build_lookup_table(int var_count, sav_ctx_t *ctx) {
-    varlookup_t *table = malloc(var_count * sizeof(varlookup_t));
+    varlookup_t *table = readstat_malloc(var_count * sizeof(varlookup_t));
     int offset = 0;
     int i;
     spss_varinfo_t *last_info = NULL;
     for (i=0; i<ctx->var_index; i++) {
-        spss_varinfo_t *info = &ctx->varinfo[i];
+        spss_varinfo_t *info = ctx->varinfo[i];
 
         if (last_info == NULL || strcmp(info->name, last_info->name) != 0) {
             varlookup_t *entry = &table[offset++];
@@ -60,7 +61,7 @@ static varlookup_t *build_lookup_table(int var_count, sav_ctx_t *ctx) {
 }
 
 
-#line 64 "src/spss/readstat_sav_parse.c"
+#line 65 "src/spss/readstat_sav_parse.c"
 static const char _sav_long_variable_parse_actions[] = {
 	0, 1, 3, 1, 5, 2, 4, 1, 
 	3, 6, 2, 0
@@ -614,7 +615,7 @@ static const int sav_long_variable_parse_start = 1;
 static const int sav_long_variable_parse_en_main = 1;
 
 
-#line 64 "src/spss/readstat_sav_parse.rl"
+#line 65 "src/spss/readstat_sav_parse.rl"
 
 
 readstat_error_t sav_parse_long_variable_names_record(void *data, int count, sav_ctx_t *ctx) {
@@ -637,11 +638,12 @@ readstat_error_t sav_parse_long_variable_names_record(void *data, int count, sav
     if (ctx->converter) {
         size_t input_len = count;
         size_t output_len = input_len * 4;
-        pe = p = output_buffer = malloc(output_len);
+        pe = p = output_buffer = readstat_malloc(output_len);
         size_t status = iconv(ctx->converter, 
                 (readstat_iconv_inbuf_t)&data, &input_len,
                 (char **)&pe, &output_len);
         if (status == (size_t)-1) {
+            free(table);
             free(output_buffer);
             return READSTAT_ERROR_PARSE;
         }
@@ -654,12 +656,12 @@ readstat_error_t sav_parse_long_variable_names_record(void *data, int count, sav
     int cs;
 
     
-#line 658 "src/spss/readstat_sav_parse.c"
+#line 660 "src/spss/readstat_sav_parse.c"
 	{
 	cs = sav_long_variable_parse_start;
 	}
 
-#line 663 "src/spss/readstat_sav_parse.c"
+#line 665 "src/spss/readstat_sav_parse.c"
 	{
 	int _klen;
 	unsigned int _trans;
@@ -734,49 +736,50 @@ _match:
 		switch ( *_acts++ )
 		{
 	case 0:
-#line 103 "src/spss/readstat_sav_parse.rl"
+#line 105 "src/spss/readstat_sav_parse.rl"
 	{
             varlookup_t *found = bsearch(temp_key, table, var_count, sizeof(varlookup_t), &compare_key_varlookup);
             if (found) {
-                memcpy(ctx->varinfo[found->index].longname, temp_val, str_len);
-                ctx->varinfo[found->index].longname[str_len] = '\0';
-            } else if (ctx->error_handler) {
-                snprintf(error_buf, sizeof(error_buf), "Failed to find %s\n", temp_key);
-                ctx->error_handler(error_buf, ctx->user_ctx);
+                spss_varinfo_t *info = ctx->varinfo[found->index];
+                memcpy(info->longname, temp_val, str_len);
+                info->longname[str_len] = '\0';
+            } else if (ctx->handle.error) {
+                snprintf(error_buf, sizeof(error_buf), "Failed to find %s", temp_key);
+                ctx->handle.error(error_buf, ctx->user_ctx);
             }
         }
 	break;
 	case 1:
-#line 114 "src/spss/readstat_sav_parse.rl"
+#line 117 "src/spss/readstat_sav_parse.rl"
 	{
             memcpy(temp_key, str_start, str_len);
             temp_key[str_len] = '\0';
         }
 	break;
 	case 2:
-#line 119 "src/spss/readstat_sav_parse.rl"
+#line 122 "src/spss/readstat_sav_parse.rl"
 	{
             memcpy(temp_val, str_start, str_len);
             temp_val[str_len] = '\0';
         }
 	break;
 	case 3:
-#line 130 "src/spss/readstat_sav_parse.rl"
+#line 133 "src/spss/readstat_sav_parse.rl"
 	{ str_start = p; }
 	break;
 	case 4:
-#line 130 "src/spss/readstat_sav_parse.rl"
+#line 133 "src/spss/readstat_sav_parse.rl"
 	{ str_len = p - str_start; }
 	break;
 	case 5:
-#line 132 "src/spss/readstat_sav_parse.rl"
+#line 135 "src/spss/readstat_sav_parse.rl"
 	{ str_start = p; }
 	break;
 	case 6:
-#line 132 "src/spss/readstat_sav_parse.rl"
+#line 135 "src/spss/readstat_sav_parse.rl"
 	{ str_len = p - str_start; }
 	break;
-#line 780 "src/spss/readstat_sav_parse.c"
+#line 783 "src/spss/readstat_sav_parse.c"
 		}
 	}
 
@@ -793,30 +796,31 @@ _again:
 	while ( __nacts-- > 0 ) {
 		switch ( *__acts++ ) {
 	case 0:
-#line 103 "src/spss/readstat_sav_parse.rl"
+#line 105 "src/spss/readstat_sav_parse.rl"
 	{
             varlookup_t *found = bsearch(temp_key, table, var_count, sizeof(varlookup_t), &compare_key_varlookup);
             if (found) {
-                memcpy(ctx->varinfo[found->index].longname, temp_val, str_len);
-                ctx->varinfo[found->index].longname[str_len] = '\0';
-            } else if (ctx->error_handler) {
-                snprintf(error_buf, sizeof(error_buf), "Failed to find %s\n", temp_key);
-                ctx->error_handler(error_buf, ctx->user_ctx);
+                spss_varinfo_t *info = ctx->varinfo[found->index];
+                memcpy(info->longname, temp_val, str_len);
+                info->longname[str_len] = '\0';
+            } else if (ctx->handle.error) {
+                snprintf(error_buf, sizeof(error_buf), "Failed to find %s", temp_key);
+                ctx->handle.error(error_buf, ctx->user_ctx);
             }
         }
 	break;
 	case 2:
-#line 119 "src/spss/readstat_sav_parse.rl"
+#line 122 "src/spss/readstat_sav_parse.rl"
 	{
             memcpy(temp_val, str_start, str_len);
             temp_val[str_len] = '\0';
         }
 	break;
 	case 6:
-#line 132 "src/spss/readstat_sav_parse.rl"
+#line 135 "src/spss/readstat_sav_parse.rl"
 	{ str_len = p - str_start; }
 	break;
-#line 820 "src/spss/readstat_sav_parse.c"
+#line 824 "src/spss/readstat_sav_parse.c"
 		}
 	}
 	}
@@ -824,14 +828,14 @@ _again:
 	_out: {}
 	}
 
-#line 140 "src/spss/readstat_sav_parse.rl"
+#line 143 "src/spss/readstat_sav_parse.rl"
 
 
     if (cs < 227|| p != pe) {
-        if (ctx->error_handler) {
-            snprintf(error_buf, sizeof(error_buf), "Error parsing string \"%s\" around byte #%ld/%d, character %c\n", 
-                    (char *)data, (long)(p - c_data), count, *p);
-            ctx->error_handler(error_buf, ctx->user_ctx);
+        if (ctx->handle.error) {
+            snprintf(error_buf, sizeof(error_buf), "Error parsing string \"%.*s\" around byte #%ld/%d, character %c", 
+                    count, (char *)data, (long)(p - c_data), count, *p);
+            ctx->handle.error(error_buf, ctx->user_ctx);
         }
         retval = READSTAT_ERROR_PARSE;
     }
@@ -848,7 +852,7 @@ _again:
 }
 
 
-#line 852 "src/spss/readstat_sav_parse.c"
+#line 856 "src/spss/readstat_sav_parse.c"
 static const char _sav_very_long_string_parse_actions[] = {
 	0, 1, 0, 1, 2, 1, 3, 2, 
 	4, 1, 2, 5, 2
@@ -956,7 +960,7 @@ static const int sav_very_long_string_parse_start = 1;
 static const int sav_very_long_string_parse_en_main = 1;
 
 
-#line 166 "src/spss/readstat_sav_parse.rl"
+#line 169 "src/spss/readstat_sav_parse.rl"
 
 
 readstat_error_t sav_parse_very_long_string_record(void *data, int count, sav_ctx_t *ctx) {
@@ -965,12 +969,12 @@ readstat_error_t sav_parse_very_long_string_record(void *data, int count, sav_ct
     readstat_error_t retval = READSTAT_OK;
 
     char temp_key[8*4+1];
-    int temp_val = 0;
+    unsigned int temp_val = 0;
     unsigned char *str_start = NULL;
     size_t str_len = 0;
 
     size_t error_buf_len = 1024 + count;
-    char *error_buf = malloc(error_buf_len);
+    char *error_buf = readstat_malloc(error_buf_len);
     unsigned char *p = NULL;
     unsigned char *pe = NULL;
 
@@ -982,13 +986,14 @@ readstat_error_t sav_parse_very_long_string_record(void *data, int count, sav_ct
         size_t input_len = count;
         size_t output_len = input_len * 4;
 
-        pe = p = output_buffer = malloc(output_len);
+        pe = p = output_buffer = readstat_malloc(output_len);
 
         size_t status = iconv(ctx->converter, 
                 (readstat_iconv_inbuf_t)&data, &input_len,
                 (char **)&pe, &output_len);
         if (status == (size_t)-1) {
             free(output_buffer);
+            free(error_buf);
             return READSTAT_ERROR_PARSE;
         }
     } else {
@@ -999,12 +1004,12 @@ readstat_error_t sav_parse_very_long_string_record(void *data, int count, sav_ct
     int cs;
     
     
-#line 1003 "src/spss/readstat_sav_parse.c"
+#line 1008 "src/spss/readstat_sav_parse.c"
 	{
 	cs = sav_very_long_string_parse_start;
 	}
 
-#line 1008 "src/spss/readstat_sav_parse.c"
+#line 1013 "src/spss/readstat_sav_parse.c"
 	{
 	int _klen;
 	unsigned int _trans;
@@ -1079,23 +1084,23 @@ _match:
 		switch ( *_acts++ )
 		{
 	case 0:
-#line 208 "src/spss/readstat_sav_parse.rl"
+#line 212 "src/spss/readstat_sav_parse.rl"
 	{
             varlookup_t *found = bsearch(temp_key, table, var_count, sizeof(varlookup_t), &compare_key_varlookup);
             if (found) {
-                ctx->varinfo[found->index].string_length = temp_val;
+                ctx->varinfo[found->index]->string_length = temp_val;
             }
         }
 	break;
 	case 1:
-#line 215 "src/spss/readstat_sav_parse.rl"
+#line 219 "src/spss/readstat_sav_parse.rl"
 	{
             memcpy(temp_key, str_start, str_len);
             temp_key[str_len] = '\0';
         }
 	break;
 	case 2:
-#line 220 "src/spss/readstat_sav_parse.rl"
+#line 224 "src/spss/readstat_sav_parse.rl"
 	{
             if ((*p) != '\0') { 
                 temp_val = 10 * temp_val + ((*p) - '0'); 
@@ -1103,18 +1108,18 @@ _match:
         }
 	break;
 	case 3:
-#line 232 "src/spss/readstat_sav_parse.rl"
+#line 236 "src/spss/readstat_sav_parse.rl"
 	{ str_start = p; }
 	break;
 	case 4:
-#line 232 "src/spss/readstat_sav_parse.rl"
+#line 236 "src/spss/readstat_sav_parse.rl"
 	{ str_len = p - str_start; }
 	break;
 	case 5:
-#line 234 "src/spss/readstat_sav_parse.rl"
+#line 238 "src/spss/readstat_sav_parse.rl"
 	{ temp_val = 0; }
 	break;
-#line 1118 "src/spss/readstat_sav_parse.c"
+#line 1123 "src/spss/readstat_sav_parse.c"
 		}
 	}
 
@@ -1127,13 +1132,14 @@ _again:
 	_out: {}
 	}
 
-#line 242 "src/spss/readstat_sav_parse.rl"
+#line 246 "src/spss/readstat_sav_parse.rl"
 
     
     if (cs < 36 || p != pe) {
-        if (ctx->error_handler) {
-            snprintf(error_buf, error_buf_len, "Parsed %ld of %ld bytes\nRemaining bytes: %s\n", (long)(p - c_data), (long)(pe - c_data), p);
-            ctx->error_handler(error_buf, ctx->user_ctx);
+        if (ctx->handle.error) {
+            snprintf(error_buf, error_buf_len, "Parsed %ld of %ld bytes. Remaining bytes: %.*s",
+                    (long)(p - c_data), (long)(pe - c_data), (int)(pe - p), p);
+            ctx->handle.error(error_buf, ctx->user_ctx);
         }
         retval = READSTAT_ERROR_PARSE;
     }
